;Proportional Printing with Word Wrap
;original PropSys routine and character set
;by Tony Samuels from Your Spectrum, November 1985
;alternate character set (1) and widths table by Paddy Coleman
;original word wrap demo by Nathan Asshantti
;alternate character set (2) and widths table by Jim Grimwood
;linewrap routine & demo by Jim Grimwood

        org  63900            ;linewrap routine

table:  equ  64072            ;address-32 of character widths table
ch_add: equ  23645
char:   defm "a$:"            ;63900
sadd:   defw $0000            ;63903(IX+0) absolute a$ start address
posn:   defw $0001            ;63905(IX+2) relative a$ current pointer
wi:     defb $00              ;63907(IX+4) current line width
wo:     defb $00              ;63908(IX+5) last full word line width
words:  defb $00              ;63909(IX+6) number of words on current line

find:   ld   bc,(ch_add)
        push bc
        ld   bc,char
        ld   (ch_add),bc
        call $28b2            ;LOOK-VARS
        pop  bc
        ld   (ch_add),bc
        push hl

        ld   ix,sadd
        ld   (ix+6),0         ;reset word count
        ld   hl,table+32      ;HL=address of width table
        ld   b,(hl)           ;get width of a space
        ld   hl,table+128     ;HL=address of "space" width mask
        ld   (hl),255
shift:  srl  (hl)             ;restore "space" width mask
        dec  b
        jr   nz,shift

        pop  hl
        inc  hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)           ;DE=LEN a$
        ex   de,hl            ;HL=LEN a$; DE=start a$
        ld   (ix+0),e
        ld   (ix+1),d
        add  hl,de            ;HL=end a$
        push hl               ;save it
        ex   de,hl            ;HL=start a$
        ld   e,(ix+2)
        ld   d,(ix+3)         ;DE=current pointer
        add  hl,de            ;HL=current a$

nextch: pop  de               ;DE=end a$
        push de
        ex   de,hl
        and  a
        sbc  hl,de            ;end a$ - current a$
        ex   de,hl
        jr   nc,nextok
        push hl               ;reached end a$
        pop  bc
        dec  bc               ;BC=end address of last char in a$
        inc  (ix+6)           ;count last word
        jr   print

nextok: ld   a,(hl)
        cp   32
        jr   nz,notspc
        push hl
        pop  bc
        dec  bc               ;BC=end address of last word on line
        ld   e,(ix+4)         ;E=wi
        ld   (ix+5),e         ;wo=E=wi
        inc  (ix+6)           ;increment word count

notspc: ld   e,a
        ld   d,0
        push hl
        ld   hl,table         ;HL=start of width table
        add  hl,de            ;HL=character position in table
        ld   a,(hl)           ;A =character width
        add  a,(ix+4)         ;calculate new line width
        ld   (ix+4),a         ;save it
        pop  hl               ;HL=current a$
        inc  hl
        jr   nc,nextch        ;checks result of "add  a,(ix+4)"

print:  pop  de               ;DE=end a$
        ld   e,(ix+0)
        ld   d,(ix+1)         ;DE=start a$
        and  a
        sbc  hl,de            ;current a$ - start a$
        ld   (ix+2),l
        ld   (ix+3),h         ;save current pointer
        push bc
        pop  hl
        sbc  hl,de            ;end of segment - start a$
        push hl
        pop  bc               ;BC="end" variable for BASIC
        ld   a,(ix+4)
        sub  (ix+5)
        ld   (ix+4),a
        ld   hl,table+128     ;check for padding before returning
        ld   a,255            ;HL=address of "space" width mask
        sub  (ix+5)           ;A =space at end of line
        dec  (ix+6)           ;(words in line)-1 = spaces in line
        jr   z,nonspc
chkspc: sub  (ix+6)           ;subtract number of spaces in line
        jr   nc,padspc        ;return when no further padding needed
        add  a,(ix+6)
nonspc: ld   (ix+6),a         ;save any remaining bit count
        ret
padspc: srl  (hl)             ;widen "space" by one pixel
        jr   chkspc

;       ORG  64104         ;character widths table   (96 bytes)
;       ORG  64200         ;character set definition (768 bytes)

        ORG  64969

ATTRP:  EQU  23693
PIXADD: EQU  $22AA
UDG:    EQU  23675
;sadd:   equ  63903

;Reserve a bit of space for some variables.
XPOS:   DEFB 00
YPOS:   DEFB 184           ;mod to access lower screen
DFADD:  DEFW 0000
PIXPOS: DEFB 00
CHRAD:  DEFW 0000
ATFLAG: DEFB 00
CRMSK:  DEFB 00
CRAD1:  DEFW 0000
CRADR:  DEFW 64200-256
WIDTH:  DEFB 00

LPRINT: defw 0000
;Stop the routine at 64985.
UNABLE: call DOABLE        ;check if LPRINT address saved
        ret  z
        jr   ENABIT+3      ;if so, restore it
DOABLE: ld   hl,LPRINT
        ld   e,(hl)
        ld   a,e
        inc  hl
        ld   d,(hl)
        add  a,d
        ret

;Start the routine at 65000.
;Reset the printing position to the top left of the screen.
INITI:  LD   HL,184*256    ;mod to access lower screen
        LD   (XPOS),HL

;Change the pointer to the LPRINT routine in the CHANS table to point to
;the proportional printing routine.
ENABIT: ld   de,DOIT0
        LD   HL,(23631)
        LD   BC,15
        ADD  HL,BC
        ld   c,(hl)
        LD   (HL),e
        INC  HL
        ld   b,(hl)
        LD   (HL),d
        call DOABLE        ;check if LPRINT address saved
        ret  nz
        ld   (hl),b        ;if not, save it
        dec  hl
        ld   (hl),c
        RET

;Save all the registers, call the printing routine, put back all the
;registers again and leap back to the operating system.
DOIT0:  PUSH HL
        PUSH BC
        PUSH DE
        PUSH AF
        ld   ix,sadd
        CALL DOIT1
        POP  AF
        POP  DE
        POP  BC
        POP  HL
        RET

;Look to see if last character was a control code 22 - the code for AT.
DOIT1:  PUSH AF
        LD   A,(ATFLAG)
        AND  A
        JR   NZ,GETXP
        POP  AF

;If the current character is an AT control code, set the ATFLG to indicate
;that the next two codes dealt with will be the X and Y positions to print.
ATCHQ:  CP   22
        JR   NZ,CRCHQ
        LD   A,255
        LD   (ATFLAG),A
        RET

;If the last character was an AT then fetch the X and Y co-ordinates and
;move to the new printing position.
GETXP:  CP   254
        JR   Z,GETYP
        POP  AF
        LD   (XPOS),A
        LD   HL,ATFLAG
        DEC  (HL)
        RET
GETYP:  POP  AF
        LD   B,A
        LD   A,184         ;mod to access lower screen
        SUB  B
        LD   (YPOS),A
        XOR  A
        LD   (ATFLAG),A
        RET

;If the current character is a Return control code, move down 8 pixels
;and open channel 2 to deal with nasty INK and PAPER control codes.
CRCHQ:  CP   13
        JR   NZ,VCHRQ
        CALL DWNCR
        LD   A,2
        CALL $1601
SKIPC:  RET

;Burp! If the character is not between 32 and 127 then print a question
;mark instead.
VCHRQ:  CP   32
        JR   C,PRNTQM
        CP   128
        JR   C,FCHR
PRNTQM: LD   A,63

;Figure out where the character's definition is stored in memory.
;       JR   FCHR
FCHR:   ld   (ix+5),a      ;save the character code
        LD   HL,0
        LD   L,A
        ADD  HL,HL
        ADD  HL,HL
        ADD  HL,HL
        EX   DE,HL
        LD   HL,(CRADR)
        ADD  HL,DE
        LD   A,(HL)

;Store the first byte of the character (the width indicator) and stick a
;zero byte in its place.
        LD   (CRMSK),A
        LD   (HL),0

;We're getting there! Load CHRAD with the address of the seventh byte of
;the character, check the character will fit on the screen and calculate
;the address in the display file where the character will be printed.
        LD   (CRAD1),HL
        LD   BC,7

;Rotate the character definition into the correct pixel position, place it
;on screen one byte at a time and in the right colour.
        ADD  HL,BC
        LD   (CHRAD),HL
        LD   A,(YPOS)
        CP   185           ;mod to access lower screen
        JR   NC,ERR5
        CALL FITCQ
        LD   BC,(XPOS)
PATCH:  LD   A,191         ;mod to access lower screen
        CALL PIXADD+2      ;mod to access lower screen
        LD   (PIXPOS),A    ;typo (was PIXPQ)
        LD   (DFADD),HL
        LD   B,8
PRNLP:  PUSH BC
        LD   HL,(CHRAD)
        LD   A,(HL)
        DEC  HL
        LD   (CHRAD),HL
        LD   L,A
        LD   A,(PIXPOS)
        AND  A
        JR   Z,PUTIT
        LD   B,A
        LD   H,0
ROTLP:  SRL  L
        RR   H
        AND  A
        DJNZ ROTLP
PUTIT:  LD   DE,(DFADD)
        LD   A,(DE)
        XOR  L
        LD   (DE),A
        CALL COLAD
        LD   A,(PIXPOS)
        AND  A
        JR   Z,PST
        inc  e             ;<--
        jr   z,pst2        ;<--
        LD   A,(DE)
        XOR  H
        LD   (DE),A
        CALL COLAD
PST2:   DEC  DE            ;<--
PST:    LD   HL,(DFADD)
        CALL ULINE
        LD   (DFADD),HL
        POP  BC
        DJNZ PRNLP

;Get the width indicator byte and put it back at the start of the character
;definition. Then move right the correct number of pixels for the character
;just printed.
        LD   A,(CRMSK)
        LD   HL,(CRAD1)
        LD   (HL),A
        LD   A,(XPOS)
        LD   B,A
        LD   A,(WIDTH)
        ADD  A,B
        LD   (XPOS),A
        RET

;Restore the width indicator byte to the first byte of the character
;definition and generate an Out of Screen error.
ERR5:   LD   A,(CRMSK)
        LD   HL,(CRAD1)
        LD   (HL),A
        RST  8
        DEFB 4

;Here's a handy routine that sets HL to point to the next pixel line up in
;the display file.
ULINE:  PUSH AF
        LD   A,H
        DEC  H
        AND  7
        JR   NZ,ENDIT
        LD   A,L
        SUB  32
        LD   L,A
        JR   C,ENDIT
        LD   A,H
        ADD  A,8
        LD   H,A
ENDIT:  POP  AF
        RET

;Check if the character to be printed will fit on this line - if it doesn't,
;move down 8 pixels and back to the left hand side of the screen.
FITCQ:  LD   a,(CRMSK)     ;A =character width mask
        ld   c,a           ;C =character width mask
        LD   B,8
        ld   a,(ix+5)      ;A =current character code
        cp   32
        jr   nz,CNTLP      ;skip pad check if not a space
        xor  a
        cp   (ix+6)
        jr   z,CNTLP       ;jump if no extra padding needed
        dec  (ix+6)
        srl  c
CNTLP:  AND  A
        SRL  C
        JR   NC,OUTIT
        DEC  B
        JR   CNTLP
OUTIT:  LD   A,B
        LD   (WIDTH),A
        LD   A,(XPOS)
        ADD  A,B
        RET  NC
DWNCR:  XOR  A
        LD   (XPOS),A
        LD   A,(YPOS)
        SUB  8
        LD   (YPOS),A
        RET

;And another useful routine - it calculates the relevant address in the
;attributes file from a given display file address and stores the value of
;ATTRP (the permanent PAPER and INK colours) in it.
COLAD:  PUSH HL
        PUSH AF
        LD   A,D
        RRC  A
        RRC  A
        RRC  A
        AND  3
        OR   $58
        LD   H,A
        LD   L,E
        LD   A,(ATTRP)
        LD   (HL),A
        POP  AF
        POP  HL
        RET

;That's all folks!
FINIS:  ;END